#!/usr/bin/env python
"""Extractor of war"""
import sys
from warcio.archiveiterator import ArchiveIterator
import tempfile

from hachoir.parser import createParser
from hachoir.core.tools import makePrintable
from hachoir.metadata import extractMetadata
from hachoir.core.i18n import initLocale
from sys import argv, stderr, exit
from os import walk, listdir
from os.path import join as path_join
from fnmatch import fnmatch
import codecs
import zipfile
from lxml import etree

OUTPUT_FILENAME = "metadata.csv"

MIME_MAP = {'application/msword' : 'doc',
            'application/vnd.ms-excel' : 'xls',
            'application/vnd.ms-powerpoint' : 'ppt',
            'application/vnd.openxmlformats-officedocument.wordprocessingml.document' : 'docx',
            'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet' : 'xlsx',
            'application/vnd.openxmlformats-officedocument.presentationml.presentation' : 'pptx'
            }

PATTERNS = list(MIME_MAP.keys())
XOFFICE_EXTS = ['docx', 'pptx', 'xlsx']


def extractXmeta(filename):
    try:
        zf = zipfile.ZipFile(filename, 'r')
    except zipfile.BadZipFile:
        return None
    try:
        s = zf.open('docProps/core.xml', 'r').read()
        root = etree.fromstring(s)
        meta = {}
        namespaces = {'w': 'http://schemas.openxmlformats.org/package/2006/metadata/core-properties'}
        for t in root:
            meta[t.tag.rsplit('}', 1)[-1]] = t.text

        s = zf.open('docProps/app.xml', 'r').read()
        root = etree.fromstring(s)
        namespaces = {'w': 'http://schemas.openxmlformats.org/officeDocument/2006/extended-properties'}
        for t in root:
            meta[t.tag.rsplit('}', 1)[-1]] = t.text
    except:
        meta = None
    return meta

XFIELDMAP = {'author' : 'creator', 'comment.Company' : 'Company', 'producer' : 'Application', 'creation_date' : 'created'}


class WarcExtractor:
    """WARC extraction class"""
    def __init__(self, warcdir, fields, outname):
        self.warcdir = warcdir
        self.outname = outname
        self.fields = fields
        self.charset = "UTF-8"
        self.total = 0
        self.invalid = 0
        self.xfields = []

        for field in self.fields:
            if field in XFIELDMAP.keys():
                self.xfields.append(XFIELDMAP[field])
            else:
                self.xfields.append(field)

    def main(self):
        output = codecs.open(self.outname, "w", self.charset)
        output.write('url|filename|' + '|'.join(self.fields)+ '\n')

        filenames = listdir(self.warcdir)
        for fname in filenames:
            fname = fname.lower()
            if fname[-4:] != 'warc' and fname[-7:] != 'warc.gz':
                continue
            print('Processing %s' % (fname))
            resp = open(path_join(self.warcdir, fname), 'rb')
            for record in ArchiveIterator(resp, arc2warc=True):
                if record.rec_type == 'warcinfo':
                    pass

                elif record.rec_type == 'response':
                    h = record.http_headers.get_header('content-type')
#                    print(h, record.rec_headers.get_header('WARC-Target-URI'))
                    url = record.rec_headers.get_header('WARC-Target-URI')
                    if h and h in PATTERNS:
                        filename = url.rsplit('?', 1)[0].rsplit('/', 1)[-1].lower()
                        s = ';'.join([url, h if h else ''])
                        print(s)
                        self.total += 1
                        line = self.processRecord(record, url, filename, mime=h)
                        if line:
                            output.write(line+'\n')
                        else:
                            self.invalid += 1
                    else:
                        filename = url.rsplit('?', 1)[0].rsplit('/', 1)[-1].lower()
                        ext = filename.rsplit('.', 1)[-1]
                        if len(ext) in [3, 4] and ext in MIME_MAP.values():
                            s = ';'.join([url, h if h else ''])
                            print(s)
                            self.total += 1
                            line = self.processRecord(record, url, filename, mime=h)
                            if line:
                                output.write(line+'\n')
                            else:
                                self.invalid += 1

                else:
                    pass
        output.close()
        self.summary()

    def summary(self):
#        print(file=stderr)
        print("Valid files: %s" % (self.total - self.invalid))
        print("Invalid files: %s" % self.invalid)
        print("Total files: %s" % self.total)
#        print(file=stderr)
        print("Result written into %s" % OUTPUT_FILENAME)

    def processRecord(self, record, url, filename, mime=None):
        if mime and mime in MIME_MAP.keys():
            ext = MIME_MAP[mime]
        else:
            ext = filename.rsplit('.', 1)[-1]
        temp = tempfile.NamedTemporaryFile(suffix=ext, dir='tmp', mode='wb', delete=False)
        temp.write(record.raw_stream.read())
        temp.close()
        print("[%s] Process file %s..." % (self.total, filename))
        if ext in XOFFICE_EXTS:
            line = [filename, url]
            meta = extractXmeta(temp.name)
            if not meta:
                return None
            for field in self.xfields:
                if field in meta.keys():
                    line.append(meta[field] if meta[field] is not None else "")
                else:
                    line.append('')
            result = '|'.join(line)
            pass
        else:
            try:
                parser = createParser(temp.name)
            except:
                print("Unable to parse file")
                return None
            if not parser:
                print("Unable to parse file")
                return None
            try:
                metadata = extractMetadata(parser)
            except Exception as err:
                print("Metadata extraction error: %s" % str(err))
                return None
            if not metadata:
                print("Unable to extract metadata")
                return None

            filename = makePrintable(filename, self.charset)
            line = [filename, url]
            temp.close()

            for field in self.fields:
                parts = field.split('.')
                if len(parts) == 1:
                    value = metadata.getText(field, '')
                    value = makePrintable(value, self.charset).strip()
                else:
                    vals = metadata.getValues(parts[0])
                    items = {}
                    for item in vals:
                        try:
                            k, v = item.split(':', 1)
                            items[k.strip()] = v.strip()
                        except:
                            pass
                    value = items.get(parts[1], '')
                line.append(value[:500].replace(';', ','))
            result = '|'.join(line)
        return result


def main():
    initLocale()
    if len(argv) != 4:
        print("usage: %s directory fields outname" % argv[0])
#        print(file=stderr)
        print("eg. %s . title,creation_date outname" % argv[0])
        exit(1)
    warcdir = argv[1]
    outname = argv[3]
    fields = [field.strip() for field in argv[2].split(",")]
    WarcExtractor(warcdir, fields, outname).main()


if __name__ == '__main__':
    main()